/*
 Logic Observer (Interface) - a CPLD Based Logic Analyser
 Version: 1.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 12/11/2013
 Last Modified: 28/02/2015
 
 Used to interface to the Logic Observer to set trigger settings, start the capture and extract data.
 Automatically converts extracted data to Sigrok Pulseview v1 data format, uses 7zip to compress files.
 
 Based on V-USB example code by Christian Starkjohann and AVR ATtiny USB Tutorial at http://codeandlife.com
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

// this is libusb, see http://libusb.sourceforge.net/ 
#include <usb.h>

// same as in main.c
#define USB_TRANSFER 1
#define USB_SETTINGS 2
#define USB_STATUS 3
#define USB_CANCEL 4

usb_dev_handle *handle = NULL;
int nBytes = 0;
char buffer[256];

uint8_t start_capture = 0;
uint8_t trigger_pin = 0xFF;
uint8_t trigger_level = 1;
uint8_t trigger_delay = 0;
uint8_t trigger_count = 1;

uint8_t sample_rate = 0;
uint8_t sample_rate_timer[8] = {0x50,      0,      1,       3,       9,      19,       39,       99};
char sample_rate_text[8][9] = {"50 MS/s", "10 MS/s", "5 MS/s", "2.5 MS/s", "1 MS/s", "500 KS/s", "250 KS/s", "100 KS/s"};

uint8_t samples = 5;
char samples_text[6][6] = {"8K", "16K", "32K", "64K", "96K", "131K"};
uint32_t samples_size[7] = {8128, 16002, 32004, 64008, 96012, 131064};

// used to get descriptor strings for device identification 
static int usbGetDescriptorString(usb_dev_handle *dev, int index, int langid, 
                                  char *buf, int buflen) {
    char buffer[256];
    int rval, i;

	// make standard request GET_DESCRIPTOR, type string and given index 
    // (e.g. dev->iProduct)
	rval = usb_control_msg(dev, 
        USB_TYPE_STANDARD | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
        USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, 
        buffer, sizeof(buffer), 1000);
        
    if(rval < 0) // error
		return rval;
	
    // rval should be bytes read, but buffer[0] contains the actual response size
	if((unsigned char)buffer[0] < rval)
		rval = (unsigned char)buffer[0]; // string is shorter than bytes read
	
	if(buffer[1] != USB_DT_STRING) // second byte is the data type
		return 0; // invalid return type
		
	// we're dealing with UTF-16LE here so actual chars is half of rval,
	// and index 0 doesn't count
	rval /= 2;
	
	// lossy conversion to ISO Latin1 
	for(i = 1; i < rval && i < buflen; i++) {
		if(buffer[2 * i + 1] == 0)
			buf[i-1] = buffer[2 * i];
		else
			buf[i-1] = '?'; // outside of ISO Latin1 range
	}
	buf[i-1] = 0;
	
	return i-1;
}

static usb_dev_handle * usbOpenDevice(int vendor, char *vendorName, 
                                      int product, char *productName) {
	struct usb_bus *bus;
	struct usb_device *dev;
	char devVendor[256], devProduct[256];
    
	usb_dev_handle * handle = NULL;
	
	usb_init();
	usb_find_busses();
	usb_find_devices();
	
	for(bus=usb_get_busses(); bus; bus=bus->next) {
		for(dev=bus->devices; dev; dev=dev->next) {			
			if(dev->descriptor.idVendor != vendor ||
               dev->descriptor.idProduct != product)
                continue;
                
            // we need to open the device in order to query strings 
            if(!(handle = usb_open(dev))) {
                fprintf(stderr, "Warning: cannot open USB device: %s\n",
                    usb_strerror());
                continue;
            }
            
            // get vendor name 
            if(usbGetDescriptorString(handle, dev->descriptor.iManufacturer, 0x0409, devVendor, sizeof(devVendor)) < 0) {
                fprintf(stderr, 
                    "Warning: cannot query manufacturer for device: %s\n", 
                    usb_strerror());
                usb_close(handle);
                continue;
            }
            
            // get product name 
            if(usbGetDescriptorString(handle, dev->descriptor.iProduct, 
               0x0409, devProduct, sizeof(devVendor)) < 0) {
                fprintf(stderr, 
                    "Warning: cannot query product for device: %s\n", 
                    usb_strerror());
                usb_close(handle);
                continue;
            }
            
            if(strcmp(devVendor, vendorName) == 0 && 
               strcmp(devProduct, productName) == 0)
                return handle;
            else
                usb_close(handle);
		}
	}
	
	return NULL;
}

// Reads the first digit of the input and ignore the rest of the input
uint8_t read_first_digit(void) {
	char c;
	uint8_t first_digit = 0;
	uint8_t number = 0;
	
	while((c = getchar()) != '\n' && c != EOF) {
		if (first_digit == 0) {
			number = c - 48;
			first_digit = 1;
		}
	}
	
	return number;
}

// Read the first character and ignore the rest of the input
uint8_t read_first_char(void) {
	char c;
	uint8_t first_char = 0;
	char first_character = '0';
	
	while((c = getchar()) != '\n' && c != EOF) {
		if (first_char == 0) {
			first_character = c;
			first_char = 1;
		}
	}
	
	return first_character;
}

// Reads x characters of input
void read_chars(char* char_array, uint8_t char_max) {
	char c;
	uint8_t char_count = 0;
	
	while((c = getchar()) != '\n' && c != EOF) {
		if (char_count < char_max) {
			char_array[char_count] = c;
			char_count++;
		}
	}
	char_array[char_count] = '\0';
}

// Transfer data from device
void extract_data(void) {
	// Get the date/time to be used for the filename
	time_t rawtime;
	struct tm* timeinfo;
	char timebuffer[80];
	
	time (&rawtime);
	timeinfo = localtime (&rawtime);
	strftime (timebuffer,80,"%Y-%m-%d_%H-%M-%S",timeinfo);
	
	
	// Create the PulseView metafile
	char metadata_p1[] = {0x5B, 0x67, 0x6C, 0x6F, 0x62, 0x61, 0x6C, 0x5D, 0x0A, 0x73, 0x69, 0x67, 0x72, 0x6F, 0x6B, 0x20, 0x76, 0x65, 0x72, 0x73, 0x69, 0x6F, 0x6E, 0x20, 0x3D, 0x20, 0x30, 0x2E, 0x32, 0x2E, 0x30, 0x0A, 0x5B, 0x64, 0x65, 0x76, 0x69, 0x63, 0x65, 0x20, 0x31, 0x5D, 0x0A, 0x64, 0x72, 0x69, 0x76, 0x65, 0x72, 0x20, 0x3D, 0x20, 0x66, 0x78, 0x32, 0x6C, 0x61, 0x66, 0x77, 0x0A, 0x63, 0x61, 0x70, 0x74, 0x75, 0x72, 0x65, 0x66, 0x69, 0x6C, 0x65, 0x20, 0x3D, 0x20, 0x6C, 0x6F, 0x67, 0x69, 0x63, 0x2D, 0x31, 0x0A, 0x75, 0x6E, 0x69, 0x74, 0x73, 0x69, 0x7A, 0x65, 0x20, 0x3D, 0x20, 0x31, 0x0A, 0x74, 0x6F, 0x74, 0x61, 0x6C, 0x20, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x73, 0x20, 0x3D, 0x20, 0x31, 0x36, 0x0A, 0x73, 0x61, 0x6D, 0x70, 0x6C, 0x65, 0x72, 0x61, 0x74, 0x65, 0x20, 0x3D, 0x20, 0x00};
	char metadata_p2[] = {0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x31, 0x20, 0x3D, 0x20, 0x31, 0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x32, 0x20, 0x3D, 0x20, 0x32, 0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x33, 0x20, 0x3D, 0x20, 0x33, 0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x34, 0x20, 0x3D, 0x20, 0x34, 0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x35, 0x20, 0x3D, 0x20, 0x35, 0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x36, 0x20, 0x3D, 0x20, 0x36, 0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x37, 0x20, 0x3D, 0x20, 0x37, 0x0A, 0x70, 0x72, 0x6F, 0x62, 0x65, 0x38, 0x20, 0x3D, 0x20, 0x38, 0x0A, 0x00};
	
	FILE *metaFile = fopen("metadata", "wb");
	if (metaFile == NULL) {
		fputs ("File error",stderr); 
		exit (1);
	}
	fwrite(metadata_p1, 1, strlen(metadata_p1), metaFile);
	fwrite(sample_rate_text[sample_rate], 1, strlen(sample_rate_text[sample_rate]), metaFile);
	fwrite(metadata_p2, 1, strlen(metadata_p2), metaFile);
	fclose(metaFile);
	
	
	// Create the PulseView version file
	FILE *versionFile = fopen("version", "wb");
	if (versionFile == NULL) {
		fputs ("File error",stderr); 
		exit (1);
	}
	fwrite("1", 1, 1, versionFile);
	fclose(versionFile);
	
	
	// Create the PulseView data file
	FILE *pFile = fopen("logic-1", "wb");
	if (pFile == NULL) {
		fputs ("File error",stderr); 
		exit (1);
	}
	
	printf("\nTransferring data...\n");
	nBytes = 1;
	
	
	// Write the progress file for the GUI to use
	int progessBarDivided = samples_size[samples] / 100;
	int progessBarCalc = 0;
	int bytesCount = 0;
	FILE *progressFile = fopen("progress.txt", "wb");
	if (progressFile == NULL) {
		fputs ("File error",stderr); 
		exit (1);
	}
	
	// Read the data
	while (nBytes > 0) { 
		nBytes = usb_control_msg(handle, 
			USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
			USB_TRANSFER, 0, 0, (char *)buffer, sizeof(buffer), 5000);
		
		if (nBytes > 0) {
			bytesCount += nBytes;
			
			// Display the progress bar
			while (bytesCount > progessBarCalc) {
				progessBarCalc += progessBarDivided;
				printf("#");
				fwrite("#", 1, 1, progressFile);
				fflush (progressFile);
			}
			
			fwrite(buffer, 1, nBytes, pFile);
		}
	}
	
	if (nBytes < 0) {
		printf("\nTransfer complete\n");
	}
	fclose(pFile);
	fclose(progressFile);
	
	// Zip up files for PulseView
	char zipcmd[150];
	sprintf(zipcmd, "7za.exe a -tzip %s.sr logic-1 metadata version", timebuffer);
	system(zipcmd);
}

// Update settings on the AVR
void update_settings(void) {
	char combined_settings[9];
	combined_settings[0] = trigger_pin;
	combined_settings[1] = trigger_level;
	combined_settings[2] = trigger_delay;
	combined_settings[3] = trigger_count;
	combined_settings[4] = sample_rate_timer[sample_rate];
	combined_settings[5] = samples;
	combined_settings[6] = start_capture;
	combined_settings[7] = '\0';
	
	nBytes = usb_control_msg(handle, 
			USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT, 
			USB_SETTINGS, 0, 0, combined_settings, 8, 5000);
}

int main(int argc, char **argv) {
	printf("insideGadgets - Logic Observer v1.0\n");
	
	handle = usbOpenDevice(0x16C0, "insidegadgets.com", 0x05DC, "LogicObserver");
	if (handle == NULL) {
		fprintf(stderr, "Could not find USB device!\n");
		exit(-1);
	}
	
	// Arguments from GUI
	if (argc >= 9 && strcmp("/update", argv[1]) == 0) {
		printf("%s ", argv[2]);
		printf("%s ", argv[3]);
		printf("%s ", argv[4]);
		printf("%s ", argv[5]);
		printf("%s ", argv[6]);
		printf("%s ", argv[7]);
		printf("%s ", argv[8]);
		
		trigger_pin = atoi(argv[2]);
		trigger_level = atoi(argv[3]);
		trigger_delay = atoi(argv[4]);
		trigger_count = atoi(argv[5]);
		sample_rate = atoi(argv[6]);
		samples = atoi(argv[7]);
		start_capture = atoi(argv[8]);
		
		update_settings();
		usb_close(handle);
		exit(1);
	}
	else if (argc == 2 && strcmp("/cancel", argv[1]) == 0) { // Cancel capture
		nBytes = usb_control_msg(handle, 
			USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
			USB_CANCEL, 0, 0, (char *)buffer, sizeof(buffer), 5000);
		
		printf("\nCancelled capture/transfer.\n");
		usb_close(handle);
		exit(1);
	}
	else if (argc == 2 && strcmp("/check", argv[1]) == 0) { // Check capture
		nBytes = usb_control_msg(handle, 
			USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
			USB_STATUS, 0, 0, (char *)buffer, sizeof(buffer), 5000);
		
		int checkCaptureComplete = atoi(buffer);
		if (checkCaptureComplete == 1) {
			printf("\n Captured");
			usb_close(handle);
			exit(2);
		}
		else {
			printf("\nWaiting");
		}
		usb_close(handle);
		exit(1);
	}
	else if (argc == 4 && strcmp("/read", argv[1]) == 0) { // Read capture
		sample_rate = atoi(argv[2]);
		samples = atoi(argv[3]);
		
		extract_data();
		usb_close(handle);
		exit(1);
	}
	
	while (1) {
		// Calculate trigger pins
		char trigger_str[12] = "";
		int x = 0;
		for (x = 0; x < 8; x++) {
			if (trigger_pin & (1<<x)) {
				char convert_int[3];
				sprintf(convert_int, "%i", x+1);
				strcat(trigger_str, convert_int);
			}
		}
		
		// Calculate trigger level
		char trigger_level_str[6] = "";
		if (trigger_level == 1) {
			strcat(trigger_level_str, "High");
		}
		else {
			strcat(trigger_level_str, "Low");
		}
		
		printf("\n\n1. Trigger pins = %s\n2. Trigger level = %s\n3. Trigger delay = %ims\n4. Trigger count = %i\n5. Sample rate = %s\n6. Samples = %s\nS. Start capture\nE. Extract data\nQ. Exit\n>", trigger_str, trigger_level_str, trigger_delay, trigger_count, sample_rate_text[(sample_rate)], samples_text[samples]);
		
		// Read input and process
		char selection = read_first_char();
		if (selection == '1') { // Trigger pins
			printf("\nEnter pins to trigger on (or enter \"a\" for all):\n>");
			
			char selection_input[10];
			read_chars(selection_input, 8); // Read input, 8 chars max
			trigger_pin = 0;
			
			// Check pins
			if (selection_input[0] == 'a') {
				trigger_pin = 255;
			}
			else {
				int x = 0;
				for (x = 0; x < strlen(selection_input); x++) {
					trigger_pin += (1<<(selection_input[x]-49));	
				}
			}
			update_settings();
		}
		else if (selection == '2') { // Trigger level
			printf("\nEnter h for High and l for Low:\n>");
			
			char selection_input[3];
			read_chars(selection_input, 2);
			if (selection_input[0] == 'h') {
				trigger_level = 1;
			}
			else {
				trigger_level = 0;
			}
			update_settings();
		}
		else if (selection == '3') { // Trigger delay
			printf("\nEnter trigger delay in milliseconds:\n>");
			
			char selection_input[7];
			read_chars(selection_input, 3); // Read input, 3 chars max
			trigger_delay = atoi(selection_input);
			
			update_settings();
		}
		else if (selection == '4') { // Trigger count
			printf("\nEnter number of triggers to wait for:\n>");
			
			char selection_input[5];
			read_chars(selection_input, 3); // Read input, 3 chars max
			trigger_count = atoi(selection_input);
			
			update_settings();
		}
		else if (selection == '5') { // Sample rate
			printf("\nSelect a sample rate:  \n1. 50 MS/s  \n2. 10 MS/s  \n3. 5 MS/s  \n4. 2.5 MS/s  \n5. 1 MS/s  \n6. 500 KS/s  \n7. 250 KS/s  \n8. 100 KS/s\n>");
			
			char selection_input[3];
			read_chars(selection_input, 2);
			sample_rate = atoi(selection_input) - 1;
			if (sample_rate == 0 || sample_rate >= 9) {
				sample_rate = 0;
			}
			update_settings();
		}
		else if (selection == '6') { // Samples
			printf("\nSelect the number of samples:  \n1. 8K  \n2. 16K  \n3. 32K  \n4. 64K  \n5. 96K  \n6. 131K\n>");
			
			char selection_input[3];
			read_chars(selection_input, 2);
			samples = atoi(selection_input) - 1;
			if (samples == 0 || sample_rate >= 7) {
				samples = 0;
			}
			update_settings();
		}
		else if (selection == 's') {
			printf("\nWaiting for trigger.");
			start_capture = 1;
			update_settings();
			start_capture = 0;
			
			uint8_t captureComplete = 0;
			while (captureComplete == 0) { // Wait until capture is complete
				nBytes = usb_control_msg(handle, 
					USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
					USB_STATUS, 0, 0, (char *)buffer, sizeof(buffer), 5000);
				
				captureComplete = atoi(buffer);
				if (captureComplete == 0) {
					printf(".");
				}
				
				Sleep(500);
			}
			
			printf(" Triggered\n");
		}
		else if (selection == 'e') {
			update_settings();
			extract_data();
		}
		else if (selection == 'q') {
			break;
		}
	}
	
	usb_close(handle);
	
	return 0;
}
